/**
 * This source file is part of OgreColladaPlugin
 * an addon for OGRE (Object-oriented Graphics Rendering Engine)
 * For the latest info, see http://www.ogre3d.org/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.

 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA, or go to
 * http://www.gnu.org/copyleft/lesser.txt.
 *
 * @authors	Philipp Hartl, Jeff Doyle (nfz)
 * @see		README
 */

#ifndef __ColladaManager_H__
#define __ColladaManager_H__

#include "OgreSingleton.h"
#include "OgreResourceManager.h"
#include "OgreColladaPrerequisites.h"
#include "OgreColladaDocument.h"

namespace Ogre {

    /** Class for managing Collada Documents for Ogre.
        @remarks
        @par
            When loaded from a XML Document, Collada objects are in an 'unloaded' state and only stores the settings
            required. It does not at that stage load any Ogre resources. This is because the Collada Documents may be
            loaded 'en masse', but only a subset will actually be required.
        @par
            Because this is a subclass of ResourceManager, any files loaded will be searched for in any path or
            archive added to the resource paths/archives. See ResourceManager for details.
        @par
            For a definition of the Collada Document format, see .
    */
    class _OgreColladaExport ColladaManager : public ResourceManager, public Singleton<ColladaManager>
    {
    public:
        ColladaManager(void);
        virtual ~ColladaManager(void);

        /** Override standard Singleton retrieval.
        @remarks
        Why do we do this? Well, it's because the Singleton
        implementation is in a .h file, which means it gets compiled
        into anybody who includes it. This is needed for the
        Singleton template to work, but we actually only want it
        compiled into the implementation of the class based on the
        Singleton, not all of them. If we don't change this, we get
        link errors when trying to use the Singleton-based class from
        an outside dll.
        @par
        This method just delegates to the template version anyway,
        but the implementation stays in this single compilation unit,
        preventing link errors.
        */
        static ColladaManager& getSingleton(void);
        /** Override standard Singleton retrieval.
        @remarks
        Why do we do this? Well, it's because the Singleton
        implementation is in a .h file, which means it gets compiled
        into anybody who includes it. This is needed for the
        Singleton template to work, but we actually only want it
        compiled into the implementation of the class based on the
        Singleton, not all of them. If we don't change this, we get
        link errors when trying to use the Singleton-based class from
        an outside dll.
        @par
        This method just delegates to the template version anyway,
        but the implementation stays in this single compilation unit,
        preventing link errors.
        */
        static ColladaManager* getSingletonPtr(void);

        /**  Load a Collada dae resource. This is a conveniance method that allows the caller to set
         *  the scenemanager to be used by the document and pass settings to be used when loading.
         * @param   daeFileName is the name of the dae file to be loaded.
         * @param   sceneMgr    is the Ogre scenemanager that that will be used when converting
         * Collada resources to Ogre resource.  Defaults to NULL.
         *
         * @param   loadParams is a list of parameters used to configure how the dae document is to be loaded.
         *                     Currently not implemented. Defaults to no parameters being passed.
         * @param   group   is the resource group to load the resource from. It defaults to DEFAULT_RESOURCE_GROUP_NAME.
        */
		ColladaDocumentPtr load(const String& daeFileName, SceneManager* sceneMgr = 0,
            NameValuePairList* loadParams = 0,
            const String& group = ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME
            );


    protected:
        /// version number of the plugin
        String mVersion;
		/// Overridden from ResourceManager

		Resource* createImpl(const String& name, ResourceHandle handle, 
			const String& group, bool isManual, ManualResourceLoader* loader,
            const NameValuePairList* params);
    };
}


#endif


